// Fast Raytracer
// Changelog 10.09.16

#ifndef __Vector__
#define __Vector__

#include <cmath>
#include <iostream>

inline double min(const double& a, const double& b) {
    return (a < b) ? a : b;
}

inline double max(const double& a, const double& b) {
    return (a > b) ? a : b;
}

//! \class Vector vector.h
//! \brief Cette classe d�finit des vecteurs et des sommets dans l'espace.

class vector3D {
protected:
    double x, y, z;
public:

    inline vector3D() : x(0), y(0), z(0) {

    }

    inline vector3D(const double& a, const double& b, const double& c) : x(a), y(b), z(c) {

    }

    inline vector3D(const vector3D& v) : x(v.x), y(v.y), z(v.z) {

    }

    // Functions to access Vector class components

    inline double & operator[] (int i) {
        if (i == 0) return x;
        else if (i == 1) return y;
        else return z;
    }

    inline const double & operator[] (int i) const {
        if (i == 0) return x;
        else if (i == 1) return y;
        else return z;
    }

    inline vector3D scale(const vector3D& b) const {
        return vector3D(x * b.x, y * b.y, z * b.z);
    }

    // Unary operators
    vector3D operator+() const;
    vector3D operator-() const;

    // Assignment operators
    vector3D& operator+=(const vector3D&);
    vector3D& operator-=(const vector3D&);
    vector3D& operator*=(const vector3D&);
    vector3D& operator/=(const vector3D&);
    vector3D& operator*=(double);
    vector3D& operator/=(double);
    vector3D& operator=(const vector3D&);

    // Binary operators
    friend vector3D operator+(const vector3D&, const vector3D&);
    friend vector3D operator-(const vector3D&, const vector3D&);

    friend double operator*(const vector3D&, const vector3D&);

    friend vector3D operator*(const vector3D&, double);
    friend vector3D operator*(double, const vector3D&);
    friend vector3D operator/(const vector3D&, double);

    friend vector3D operator/(const vector3D&, const vector3D&);

    // Boolean functions
    friend int operator==(const vector3D&, const vector3D&);
    friend int operator!=(const vector3D&, const vector3D&);
    friend int operator<(const vector3D&, const vector3D&);
    friend int operator<=(const vector3D&, const vector3D&);
    friend int operator>(const vector3D&, const vector3D&);
    friend int operator>=(const vector3D&, const vector3D&);
    friend vector3D min(const vector3D&, const vector3D&);
    friend vector3D max(const vector3D&, const vector3D&);

    friend vector3D Orthogonal(const vector3D&);

    // Norm
    friend double Norm(const vector3D&);
    friend vector3D Normalized(const vector3D&);
    
    friend std::ostream & operator <<(std::ostream &, vector3D const &);
};

// Unary operators

inline vector3D vector3D::operator+() const {
    return *this;
}

inline vector3D vector3D::operator-() const {
    return vector3D(-x, -y, -z);
}

// Assignment unary operators

inline vector3D& vector3D::operator+=(const vector3D& u) {
    x += u.x;
    y += u.y;
    z += u.z;
    return *this;
}

inline vector3D& vector3D::operator-=(const vector3D& u) {
    x -= u.x;
    y -= u.y;
    z -= u.z;
    return *this;
}

inline vector3D& vector3D::operator*=(double a) {
    x *= a;
    y *= a;
    z *= a;
    return *this;
}

inline vector3D& vector3D::operator/=(double a) {
    x /= a;
    y /= a;
    z /= a;
    return *this;
}

inline vector3D& vector3D::operator*=(const vector3D& u) {
    x *= u.x;
    y *= u.y;
    z *= u.z;
    return *this;
}

inline vector3D& vector3D::operator/=(const vector3D& u) {
    x /= u.x;
    y /= u.y;
    z /= u.z;
    return *this;
}

inline vector3D& vector3D::operator=(const vector3D& u) {
    x = u.x;
    y = u.y;
    z = u.z;
    return *this;
}

// Binary operators

inline vector3D operator+(const vector3D& u, const vector3D& v) {
    return vector3D(u.x + v.x, u.y + v.y, u.z + v.z);
}

inline vector3D operator-(const vector3D& u, const vector3D& v) {
    return vector3D(u.x - v.x, u.y - v.y, u.z - v.z);
}

// Scalar (= Dot) product

inline double operator*(const vector3D& u, const vector3D& v) {
    return (u.x * v.x + u.y * v.y + u.z * v.z);
}

inline vector3D operator*(const vector3D& u, double a) {
    return vector3D(u.x * a, u.y * a, u.z * a);
}

inline vector3D operator*(double a, const vector3D& v) {
    return v * a;
}

// Cross product

inline vector3D operator/(const vector3D& u, const vector3D& v) {
    return vector3D(u.y * v.z - u.z * v.y, u.z * v.x - u.x * v.z, u.x * v.y - u.y * v.x);
}

inline vector3D operator/(const vector3D& u, double a) {
    return vector3D(u.x / a, u.y / a, u.z / a);
}

// Boolean functions

inline int operator==(const vector3D& u, const vector3D& v) {
    return ((u.x == v.x) && (u.y == v.y) && (u.z == v.z));
}

inline int operator!=(const vector3D& u, const vector3D& v) {
    return (!(u == v));
}

inline int operator<(const vector3D& a, const vector3D& b) {
    return ((a.x < b.x) && (a.y < b.y) && (a.z < b.z));
}

inline int operator<=(const vector3D& a, const vector3D& b) {
    return ((a.x <= b.x) && (a.y <= b.y) && (a.z <= b.z));
}

inline int operator>(const vector3D& a, const vector3D& b) {
    return ((a.x > b.x) && (a.y > b.y) && (a.z > b.z));
}

inline int operator>=(const vector3D& a, const vector3D& b) {
    return ((a.x >= b.x) && (a.y >= b.y) && (a.z >= b.z));
}

/*!
\brief Compute the Euclidean norm of a vector.
 */
inline double Norm(const vector3D& u) {
    return sqrt(u.x * u.x + u.y * u.y + u.z * u.z);
}

/*!
\brief Compute the Euclidean norm of a vector.
 */
inline vector3D Normalized(const vector3D& u) {
    return u / sqrt(u.x * u.x + u.y * u.y + u.z * u.z);
}

/*!
\brief Return a new vector with coordinates set to the minimum coordinates
of the two argument vectors.
 */
inline vector3D min(const vector3D& a, const vector3D& b) {
    return vector3D(a[0] < b[0] ? a[0] : b[0], a[1] < b[1] ? a[1] : b[1], a[2] < b[2] ? a[2] : b[2]);
}

/*!
\brief Return a new vector with coordinates set to the maximum coordinates
of the two argument vectors.
 */
inline vector3D max(const vector3D& a, const vector3D& b) {
    return vector3D(a[0] > b[0] ? a[0] : b[0], a[1] > b[1] ? a[1] : b[1], a[2] > b[2] ? a[2] : b[2]);
}

/*!
\brief Returns a new vector orthogonal to the argument vector.
 */
inline vector3D Orthogonal(const vector3D& u) {
    vector3D a = vector3D(fabs(u[0]), fabs(u[1]), fabs(u[2]));
    int i = 0;
    int j = 1;
    if (a[0] > a[1]) {
        if (a[2] > a[1]) {
            j = 2;
        }
    } else {
        i = 1;
        j = 2;
        if (a[0] > a[2]) {
            j = 0;
        }
    }
    a = vector3D(0.0, 0.0, 0.0);
    a[i] = u[j];
    a[j] = -u[i];
    return a;
}

inline std::ostream & operator <<(std::ostream & os, vector3D const & vector) {
    return os << "v(" << vector.x << ", " << vector.y << ", " << vector.z << ")";
}

#endif

